#ifndef __CLIENT_H__
#define __CLIENT_H__

#include "exco.h"
#include "task.h"
#include "communication.h"
#include "action_map.h"
#include "client_actions.h"
#include "client_reactions.h"
#include "renderer.h"
#include "event.h"
#include "boost/shared_ptr.hpp"
#include "boost/utility.hpp"
#include "rendering/render_terrain_hook.h"

#include <iostream>

namespace te
{
//#############################################################################
class client : public ec::multi_task
{
    class processCommands;
    friend class processCommands;

    class fetchInput;
    friend class fetchInput;

    class render_area;
    friend class render_area;

    class dispatchCommands;
    friend class dispatchCommands;

    friend class reaction_set_map_size;
    friend class reaction_leaf_send;
    friend class reaction_date_and_time;

public: //---------------------------------------------------------------------
    client (ec::exco& its_list,
        commLink& its_in_comm,
        commLink& its_out_comm,
        size_t userCount,
        terrain_ptr& its_terrain)
        : ec::multi_task(userCount,its_list),
          my_in_comm_(its_in_comm),
          my_out_comm_(its_out_comm),
          user_count_(userCount),
          my_task_list_(userCount),
          my_renderer_(),
          my_reactions_(std::string("")),
          my_actions_(std::string("")),
          my_terrain_(its_terrain)
{
    my_terrain_ = boost::shared_ptr<terrain<tileData> >\
            (new terrain<tileData>(32, 32));

    my_area_renderer_ = ec::task_ptr(new render_area(*this));
    render_area& tmp = *(static_cast<render_area*>(my_area_renderer_.get()));
    tmp.resize(4,4);

    my_task_list_.insert(ec::task_ptr(new processCommands (*this)));
    my_task_list_.insert(ec::task_ptr(new fetchInput      (*this)));
    my_task_list_.insert(my_area_renderer_);
    my_task_list_.insert(ec::task_ptr(new dispatchCommands(*this)));
    
    my_reactions_.associate(dummy        ,reaction_ptr (new action_dummy          (     )));
    my_reactions_.associate(set_map_size ,reaction_ptr (new reaction_set_map_size (*this)));
    my_reactions_.associate(leaf_send    ,reaction_ptr (new reaction_leaf_send    (*this)));
    my_reactions_.associate(date_and_time,reaction_ptr (new reaction_date_and_time(*this)));

    my_actions_.associate(dummy   ,action_ptr (new action_dummy()));
    my_actions_.associate(exit    ,action_ptr (new action_exit(my_out_comm_)));
    my_actions_.associate(map_make,action_ptr (new action_map_make(my_out_comm_)));
    my_actions_.associate(area_req,action_ptr (new action_area_req(my_out_comm_)));
    
    my_event_ = boost::shared_ptr<eventReader2>\
                       (new eventReader2(my_actions_,\
                       my_renderer_,\
                       my_renderer_.winHandle(),\
                       my_renderer_.winWidth(),\
                       my_renderer_.winHeight()));

}
//-----------------------------------------------------------------------------
void logic()
{
    ec::threadlessExecutor tx(my_task_list_,"");
    tx.go();
    my_task_list_.refresh();
}
private: //--------------------------------------------------------------------
    //#########################################################################
    //this task will fetch and process all commands in the input comm queue
    class processCommands : public ec::single_task
    {
    public: //-----------------------------------------------------------------
        processCommands(client& its_client) \
            : ec::single_task(its_client.my_task_list_),
              my_client_     (its_client)
        {
        };
        //---------------------------------------------------------------------
        void logic()
        {
            //std::cout << "client::processDatagrams\n";
            te::datagram dgram;
            while (my_client_.my_in_comm_.fetch(dgram))
            {
                my_client_.my_reactions_[dgram.command_]->setParams(dgram);
                my_client_.my_reactions_[dgram.command_]->invoke();
            }
        };
    private:
        client& my_client_;
    };
    //#########################################################################
    //this task will fetch and process all input from mouse and keyboard
    class fetchInput : public ec::single_task
    {
    public: //-----------------------------------------------------------------
        fetchInput(client& its_client) \
            : ec::single_task (its_client.my_task_list_),
              my_client_      (its_client)
        {
        };
        //---------------------------------------------------------------------
        void logic()
        {
            my_client_.my_renderer_();
            (*my_client_.my_event_)();

            my_client_.my_current_long_ = 4;
            my_client_.my_current_latt_ = 4;
            my_client_.my_renderer_.centerCamera(my_client_.my_current_long_, my_client_.my_current_latt_);

            my_camera_fov_ = my_client_.my_renderer_.getCameraFov();

            if (my_camera_fov_ >= 2.59)
                my_client_.my_current_viewable_radius_ = 1;
            if (my_camera_fov_ >= 2.69)
                my_client_.my_current_viewable_radius_ = 2;
            if (my_camera_fov_ >= 2.79)
                my_client_.my_current_viewable_radius_ = 3;
            if (my_camera_fov_ >= 2.89)
                my_client_.my_current_viewable_radius_ = 4;
            if (my_camera_fov_ >= 2.99)
                my_client_.my_current_viewable_radius_ = 5;

        };
    private:
        client& my_client_;
        float my_camera_fov_;
    };
    //#########################################################################
    class render_area : public ec::multi_task
    {
        class leaf_drawer;
        friend class leaf_drawer;
    public: //-----------------------------------------------------------------
        /**
        @remark
            constructor
        @param
            its_list - exco render_area is inside of
            userCount - maximum number of executors which will
                        enter render_area
        */
    render_area(client& its_client) :
               ec::multi_task(its_client.user_count_,its_client.my_task_list_),
                   my_client_(its_client),
             my_display_list_(200,its_client.my_renderer_),
                  my_barrier_(unsigned(its_client.user_count_)),
                my_task_list_(its_client.user_count_),
                my_longitude_(its_client.my_terrain_->lngt())
    {
    };
    //-------------------------------------------------------------------------
    ~render_area()
    {
        my_hooks_.clear();
    };
    //-------------------------------------------------------------------------
        /**
        @remark
            logic of the task - all executors are created here,
            tasks are fired up etc
        */
    void logic()
    {
        if (my_barrier_.wait())
        {
            // 1 - get the bounds of the viewable area
            int lllo = int(my_client_.my_current_long_ -\
                my_client_.my_current_viewable_radius_);
                
            int llla = int(my_client_.my_current_latt_ -\
                my_client_.my_current_viewable_radius_);
                
            int rulo = int(my_client_.my_current_long_ +\
                my_client_.my_current_viewable_radius_);
                
            int rula = int(my_client_.my_current_latt_ +\
                my_client_.my_current_viewable_radius_);

            ensure<int> (0,int(my_client_.my_terrain_->lngt())-1,lllo);
            ensure<int> (0,int(my_client_.my_terrain_->latt())-1,llla);
            ensure<int> (0,int(my_client_.my_terrain_->lngt())-1,rulo);
            ensure<int> (0,int(my_client_.my_terrain_->latt())-1,rula);

            // 2 - start inserting tasks
            for (leaf_size_t x = leaf_size_t(lllo);x <= leaf_size_t(rulo);++x)
            {
                for (leaf_size_t y = leaf_size_t(llla);y <= leaf_size_t(rula);++y)
                {
                    my_task_list_.insert(ec::task_ptr(new leaf_drawer(my_hooks_[x*my_longitude_  + y],my_task_list_)));
                }
            }
            // done - now create executors
        }
        my_barrier_.wait();
        ec::threadlessExecutor thex(my_task_list_,"");
        thex.go();
        if(my_barrier_.wait())
        {            
            // 3 - clean the exco
            my_task_list_.clear();
        };
    }
    //-------------------------------------------------------------------------
        /**
        * @remark
        *      this method is called when the map size has changed and thus new
        *      set of hooks is needed
        *      this function is NOT reentrant! It is only called from
        *      service of map resize task
        * @param
        *      longitunal and lattitunal size in leafs of resized terrain
        */
    void resize(leaf_size_t longitude, leaf_size_t lattitude)
    {
        //std::cout << unsigned(my_hooks_.size());
        my_hooks_.clear();
        te::terrain<tileData>& terrain = *(my_client_.my_terrain_.get());
        for (size_t x = 0;x < longitude;++x)
        {
            for(size_t y = 0;y < lattitude;++y)
            {
                my_hooks_.push_back(hook_ptr(new render_terrain_hook(terrain.getLeaf(x,y),\
                                                                     my_client_.my_renderer_,\
                                                                     my_display_list_,\
                                                                     x,y,my_client_.my_dayOfYear_,
                                                                     longitude)));
            }
        }
        my_longitude_ = longitude;
        my_display_list_.flush();
        //std::cout << unsigned(my_hooks_.size());
    }
    private: //----------------------------------------------------------------
        //#####################################################################
            /**
            @remark
                this class is a task that will process a leaf of map.
                this objects are created and destroyed on each frame
            */
        class leaf_drawer : public ec::single_task
        {
        public: //-------------------------------------------------------------
            leaf_drawer(hook_ptr& its_hook, ec::exco& its_list) :
                ec::single_task(its_list),  my_hook_(its_hook)
            {
            }
            void logic()
            {
                my_hook_->build();
            }
        private: //------------------------------------------------------------
            hook_ptr& my_hook_;
        };
        //#####################################################################
    //-------------------------------------------------------------------------
                          std::vector<hook_ptr> my_hooks_;
                    render_terrain_display_list my_display_list_;
                                        client& my_client_;
                                 boost::barrier my_barrier_;
                                       ec::exco my_task_list_;
                                    leaf_size_t my_longitude_;
    };
    //#########################################################################
    // this task dispatches commands for the server
    class dispatchCommands : public ec::block_task
    {
    public: //-----------------------------------------------------------------
        dispatchCommands(client& its_client) \
            : ec::block_task (its_client.user_count_, its_client.my_task_list_),
              my_client_     (its_client)
        {
        };
        //---------------------------------------------------------------------

#define AAR_PTR static_cast<action_area_req*> \
    (my_client_.my_actions_[area_req].get())

        void logic()
        {
            int lllo = int(my_client_.my_current_long_ -\
                my_client_.my_current_viewable_radius_);
            
            int llla = int(my_client_.my_current_latt_ -\
                my_client_.my_current_viewable_radius_);
            
            int rulo = int(my_client_.my_current_long_ +\
                my_client_.my_current_viewable_radius_);
            
            int rula = int(my_client_.my_current_latt_ +\
                my_client_.my_current_viewable_radius_);

            ensure<int> (0,int(my_client_.my_terrain_->lngt())-1,lllo);
            ensure<int> (0,int(my_client_.my_terrain_->latt())-1,llla);
            ensure<int> (0,int(my_client_.my_terrain_->lngt())-1,rulo);
            ensure<int> (0,int(my_client_.my_terrain_->latt())-1,rula);

            AAR_PTR->setArea(lllo,llla,rulo,rula);

            for (;lllo <= rulo;++lllo)
            {
                for(int tmp = llla;tmp <= rula;++tmp)
                {
                    AAR_PTR->addVersion((*my_client_.my_terrain_).getVersionOf(lllo,tmp));
                }
            }

            my_client_.my_actions_[area_req]->invoke();
        };
    private:
        client& my_client_;
    };

#undef AAR_PTR

    //#########################################################################
                                        commLink& my_in_comm_;
                                        commLink& my_out_comm_;
                                         ec::exco my_task_list_;
                                           size_t user_count_;
                                         renderer my_renderer_;
                  boost::shared_ptr<eventReader2> my_event_;
                                 te::reaction_map my_reactions_;
                                   te::action_map my_actions_;
                                         //- metrices -//
                                      leaf_size_t my_current_long_; // which tile the frustum
                                      leaf_size_t my_current_latt_; // centre is now on
                                      leaf_size_t my_current_viewable_radius_; // how many leafs around the centre are visible
                                         //- terrain related -//
                                     ec::task_ptr my_area_renderer_;
                                     terrain_ptr& my_terrain_;
                                        //- time metricies -//
                                           size_t my_year_;
                                           size_t my_dayOfMonth_;
                                           size_t my_dayOfYear_;
                                           size_t my_month_;
};
//#############################################################################
};

#endif
