/*
* Adam Kadlubek 2006, 2007
* Josef Drexler 2006
* Terrain Generator for TTD and Transport Empire
* under GPL license
*
* This is actually a refactor of a C++ class
* that is why all ex-methods have the explicit
* 'this_' pointer
*
* Note - to allow use in TE and TTD a set of macros is provided
* 
* result of all is by default stored in *this_
*/

#ifndef __GRIDARRAY_H__
#define __GRIDARRAY_H__

#define TE

#ifdef TE
#undef TTD
#endif

#ifdef TTD
#undef TE
#endif

typedef unsigned long ulong;
typedef unsigned int uint;

#ifdef TTD
    #include <stdint.h>
    #include <stdio.h>
    #include <stdlib.h>
    #define STATIC static
#endif

#ifdef TE
    typedef unsigned int uint32_t;
    #define STATIC
    #include "boost/thread/mutex.hpp"
    #include <cstdio>
    #include <cstdlib>
    #include "../random.h"
    inline void memcompact(){}
#endif

typedef struct {
	float** data; /* data array */
    ulong size_x; /* size of the array */
    ulong size_y; /* size of the array */

#ifdef TE
    /* used for several TE thread synchro */
    volatile ulong thSpeci1;
    volatile ulong thSpeci2;
    boost::mutex my_mutex_;
#endif

} gridArray, *gridArrayPtr;

void snipArray (gridArrayPtr source_,ulong  size_x_,ulong  size_y_,gridArrayPtr* this_);
/* copy constructor for creating the array from another
   array with its sub-cutted elements */

void makeArray (ulong size_x_, ulong size_y_,uint32_t (*randomizer)(void), gridArrayPtr* this_);
/* construct an array of a given size */

void destroyArray(gridArrayPtr* this_);
/* destructor of the array */

void mulArray(gridArrayPtr source_,gridArrayPtr this_);
/* multiply elements of one array with elements of this_ array*/

void mulArrayStrip(ulong pos, gridArrayPtr source_, gridArrayPtr this_);
/*multiply a strip of source_ by a strip of this_, store result in this_*/

void mulScalar (float scalar, gridArrayPtr this_);
/* scalar multiply operator */

void mulScalarStrip(ulong pos, float scalar, gridArrayPtr this_);
/*multiply a strip of source_ by scalar, store result in this_*/

void addArray(gridArrayPtr source_, gridArrayPtr this_);
/* addition of two arrays */

void addArrayStrip(ulong pos, gridArrayPtr source_, gridArrayPtr this_);
/*add a strip of source_ by to strip of this_, store result in this_*/

void addScalar(float scalar, gridArrayPtr this_);
/* addition of a value to all array elements */

void addScalarStrip(ulong pos, float scalar, gridArrayPtr this_);
/*add a strip of source_ to scalar this_, store result in this_*/

void scale(ulong newSize_x, ulong newSize_y, gridArrayPtr* this_);
/* scale the given array to a new size */

void filter(ulong iterations, gridArrayPtr* this_);
/* smooth the given array by x times using bilinear filtering */

void filterStrip(ulong pos, gridArrayPtr from_, gridArrayPtr this_);
/* filter bilinearily array of from_ store result in this_*/

void go(ulong iterations, gridArrayPtr* this_);
/* do the generation using the iterations set by the user */

void normalize(float hi, gridArrayPtr this_);
/* normalize the array to the low-hi range */

void constrainMap(ulong cutDown, ulong cutUp, ulong tileRange, gridArrayPtr this_);
/* change the map to fulfill the Transport
   Tycoon requirements for the vertex data */

void ttDesert(gridArrayPtr* this_);
/* make TT desert */

void print (gridArrayPtr this_);
/* print the contents of the array */

void shift_x (int dir, ulong row, gridArrayPtr this_);
/* move values by 1 on thr x axis */

void shift_y (int dir, ulong row, gridArrayPtr this_);
/* move values by q on the y axis */

/* in both cases - if dir is negative, shift low->hi else shift hi->lo */

/*stomp a circle of normal terrain to make a shoreline green
for desert creation */
void shoreStomp(ulong i_x, ulong i_y,gridArrayPtr geometry, gridArrayPtr this_);

/* cellular automata to create deserts */
void cellularAutomata(ulong pop1, ulong pop2, ulong iters, gridArrayPtr* this_);

/* decision to make a desert */
void infest(gridArrayPtr source, ulong i_x, ulong i_y, gridArrayPtr this_);

/* return size of the array */

STATIC inline ulong size_x(gridArrayPtr this_)
{
  return this_->size_x;
}

/* likewise */

STATIC inline ulong size_y(gridArrayPtr this_)
{
  return this_->size_y;
}

/* return value under indices */

STATIC inline float val (ulong i_x, ulong i_y, gridArrayPtr this_)
{

  if (size_x(this_) <= i_x || size_y(this_) <= i_y)
    return 0.0;

  return this_->data[i_x][i_y];
}

/* insert value under indices */

STATIC inline void insert (float val,ulong i_x, ulong i_y, gridArrayPtr this_)
{
  if (size_x(this_) > i_x && size_y(this_) > i_y)
    this_->data[i_x][i_y] = val;
}

/* make sure value differs no more than 1 from neighbour */

STATIC inline void adjust (ulong i_x, ulong i_y, ulong j_x, ulong j_y, int raise, ulong range, gridArrayPtr this_)
{
  ulong current = ulong(val(i_x,i_y,this_));
  ulong prev    = ulong(val(j_x,j_y,this_));

  if (raise) {

    if (current > prev + range)
      insert (float(current - range), j_x, j_y, this_);

    if (current + range < prev)
      insert (float(prev    - range), i_x, i_y, this_);

  } else {

    if (current > prev + range)
      insert (float(prev    + range), i_x, i_y, this_);

    if (current + range < prev)
      insert (float(current + range), j_x, j_y, this_);
  }
}

/* check if two verices are close/far away. needed for scaling */

STATIC inline int dist(ulong otherSize, ulong other, ulong mySize, ulong indice, gridArrayPtr this_) {

  float len = 0.0;
  float my = indice * 1.0f /mySize;
  float ot = other  * 1.0f /otherSize;

  len = (ot-my)*mySize*1.0f;
  return (len < 1.0f);
}

/* normalize a single point on the map */

STATIC static void normalizePoint(ulong i_x, ulong i_y, float low, float hi, float range, gridArrayPtr this_) 
{
    insert((val(i_x,i_y,this_)-low)*(range/hi),i_x,i_y,this_);
}

/* create a desert below certain level */

STATIC static void ttDesertBelow(int level, gridArrayPtr source, gridArrayPtr* this_)
{
    ulong i_x = 0;
    ulong i_y = 0;
    ulong limit_x = size_x(source);
    ulong limit_y = size_y(source);
    
    makeArray(limit_x, limit_y, NULL, this_);
    mulScalar(0.0f,*this_);
       
    for (i_x = 0;i_x < limit_x;++i_x)
    {
        for (i_y = 0;i_y < limit_y;++i_y)
        {
            if (val(i_x,i_y,source) > level)
            {
              insert(2.0f,i_x,i_y,*this_);
            } else if (val(i_x,i_y,source) > 0 && val(i_x,i_y,source) < level-2)
            {
              insert(1.0f,i_x,i_y,*this_);
            }
        }
    }
    
    for (i_x = 0;i_x < limit_x;++i_x)
      {
      for (i_y = 0;i_y < limit_y;++i_y)
        {
          shoreStomp(i_x,i_y,source,*this_);
        }
      }
}

#endif
