#ifndef __EXCO_H__
#define __EXCO_H__

#include "boost/thread/thread.hpp"
#include "boost/thread/mutex.hpp"
#include "boost/thread/barrier.hpp"
#include "boost/shared_ptr.hpp"
#include "boost/lexical_cast.hpp"
#include "boost/utility.hpp"

#include <string>
#include <iostream>
#include <list>

namespace ec
{
class exco;
//#############################################################################
class task
{
public: //---------------------------------------------------------------------
    /**
    @remarks
        constructor
    @param
        list in which this task will be put. May be used in internal
        work of a task
    */
    task(exco& its_list);
//-----------------------------------------------------------------------------
    virtual ~task();
//-----------------------------------------------------------------------------
    /**
    @remarks
        called by each executor when entering a task
        specialized in single/block/multi tasks
    @param
        counter towards list finish, should be increased if task
        is completed, otherwise should be intact
    */
    virtual void invoke (volatile size_t& finishCount) = 0;
//-----------------------------------------------------------------------------
    /**
    @remarks
        checks if a task is obsolete. Obsolete tasks are
        periodically removed from exco
    @return
        'true' if task is obsolete,'false' otherwise
    */
    virtual bool isObsolete() = 0;          
//-----------------------------------------------------------------------------
    /**
    @remarks
        check if a task is finished. Exco is finished when all tasks are\
        finished
    @return
        'true' if a task is finished,'false' otherwise
    */
    virtual bool isFinished() = 0;  
//-----------------------------------------------------------------------------
    /**
    @remarks
        put this task into 'fresh' state
    @param
        include obsolete tasks
    */
    virtual void refresh(bool also_obsolete) = 0;
protected: //------------------------------------------------------------------
    exco& my_list_;
};
//#############################################################################
typedef boost::shared_ptr<task> task_ptr;
//#############################################################################
    /**
    @remarks
        main execution container class, see documentation for details
    */
class exco
{
private: //--------------------------------------------------------------------
    typedef std::list<task_ptr>                     excoList;
    typedef std::list<task_ptr>::iterator           excoListIterator;
public: //---------------------------------------------------------------------
    //#########################################################################
        /**
        @remarks
            iterator of exco
        */
    class excoIterator
    {
    public: //-----------------------------------------------------------------
        friend class exco;
        /**
        @remarks
            constructor
        */
        excoIterator();
        /**
        @remarks
            put iterator on one position forward
        */
        void operator++();
    private: //----------------------------------------------------------------
        excoListIterator iter;
                    bool isInitialized_;
    };
    //#########################################################################
//-----------------------------------------------------------------------------
    /**
    @remarks
        constructor
    @param
        precise amount of executors which will be traversing this exco
    */
    exco(size_t numOfUsers);
//-----------------------------------------------------------------------------
    ~exco();
//-----------------------------------------------------------------------------
    /**
    @remarks
        insert one task to the list, note you need to call 'merge' to make this
        task usable. 'merge' should be called before/after processing of exco.
        'merge' is called automatically when exco finishes processing
    @param
        task_ptr to a task to be inserted
    */
    void insert(const task_ptr& task);
//-----------------------------------------------------------------------------
    /**
    @remarks
        called by an executor, attempt to obtain and execute one task
    @param
        iterator to use to traverse the list
    @return
        'false' if exco is finished, 'true' otherwise
    */
    bool execute(excoIterator& iter);
//-----------------------------------------------------------------------------
    /**
    @remarks
        delete all obsolete tasks. NOT THREAD SAFE!. Call only when exco is
        NOT being processed
    */
    void flush();
//-----------------------------------------------------------------------------
    /**
    @remarks
        delete all tasks
    */
    void clear();
//-----------------------------------------------------------------------------
    /**
    @remarks
        refresh all tasks
    */
    bool refresh();
private: //--------------------------------------------------------------------
    /**
    @remarks
        merge inserted tasks with main processing list
    */
    void merge();
//-----------------------------------------------------------------------------
    /**
    @remarks
        check if the exco is completed
    */
    bool isCompleted();
//-----------------------------------------------------------------------------
    volatile size_t entryCount_;
    volatile size_t finishedCount_;
    volatile bool   isInitialized_;
    size_t          numOfUsers_;
    excoList        tasks;
    excoList        tail;
    boost::mutex    insert_mtx_;
    boost::barrier  service_barr_;
};
//#############################################################################
class executor : private boost::noncopyable
{
public: //---------------------------------------------------------------------
    class delegate;
    friend class delegate;
public: //---------------------------------------------------------------------
    /**
    @remarks
        constructor
    @param
        list targeted by this executor
    @param
        string ID of this executor
    */
    executor(exco& target, std::string id);
//-----------------------------------------------------------------------------
    ~executor();
//-----------------------------------------------------------------------------
    /**
    @remarks
        check if the processing has finished
    */
    bool isFinished();
public: //---------------------------------------------------------------------
    //#########################################################################
    /**
    @remarks
        executor delegate, this part is buried in boost thread
    */
    class delegate
    {
    public:
    //-------------------------------------------------------------------------
        /**
        @remarks
            constructor
        @param
            parent/owner of this delegate
        */
        delegate(executor& parent);
    //-------------------------------------------------------------------------
        /**
        @remark
            copy constructor, needed to avoid losing contact
            with a parent
        @param
            source delegate
        */
        delegate(const delegate& src);
    //-------------------------------------------------------------------------
        /**
        @remark - main thread function
        */
        int operator()();
    private:
    //-------------------------------------------------------------------------
        delegate& operator=(const delegate& src);
        executor&       parent_;
        volatile bool   isFinished_;
    };
    //#########################################################################
private: //--------------------------------------------------------------------

    exco&                           my_list_;
    delegate*                       my_delegate_;
    exco::excoIterator              my_iterator_;
    std::string                     id_;
    boost::shared_ptr<boost::thread>thr_;
    volatile bool                   isFinished_;
};
//#############################################################################
/**
@remark
    threadless executor, needed for all those situations when main/current
    thread has to traverse exco
*/
class threadlessExecutor : private boost::noncopyable
{
public: //---------------------------------------------------------------------
    /**
    @remarks
        constuctor
    @param
        targeted execution list
    @param
        identification string
    */
    threadlessExecutor(exco& target, std::string id);
//-----------------------------------------------------------------------------
    ~threadlessExecutor();
//-----------------------------------------------------------------------------
    /**
    @remarks
        start processing the exco
    */
    int go();
private: //--------------------------------------------------------------------
    exco&                           my_list_;
    exco::excoIterator              my_iterator_;
    std::string                     id_;
};
//#############################################################################
};

#endif