
/*
This file is a part of TRANSPORT EMPIRE
2006-2007 Copyright by Adam Kadlubek.

This file is licenced under GPL General Public Licence
*/

#ifndef __MAP_H__
#define __MAP_H__

#include <vector>
#include <boost/shared_ptr.hpp>
#include <iostream>
#include "services.h"

namespace te
{
//#############################################################################
/*
* The map in TE is composed of tiles. Each tile has four corners and height.
* Corners can differ by 15 units in 1 unit increments. 1 unit = 0.5 metres
* At least one corner value must be 0
*
* The tiles are in clockwise orientation with lower left corner
* being the lowest
*
* 2-3
* | |
* 1_4
*
* the division to triangles is highest-first, that is, the highest corner
* is the origin of the triange division
*
* coding
* bit  0-3  - 1st corner height
* bit  4-7  - 2nd corner height
* bit  8-11 - 3rd corner height
* bit 12-15 - 4th corner height
* bit 16-27 - height of a tile  (4096 values)
* bit 28-63 - flags
*	      bit 28 - tile is visible
*		  bit 29-31 - reserved
* bit 32-37 - tile's owner ID -> 64 possible owners
* bit 38-43 - tile's type -> up to 64 possible types
* bit 44-47 - vegetation density  (8 values)
* bit 48-63 - reserved
* the map is orientated in such way
*
*
*  ^ lattitude
*  |
*  |
*  |
* (0,0)---> longitude
*
*/
//#############################################################################
typedef size_t leaf_size_t;
typedef size_t tile_size_t;
//#############################################################################
class tileData
{
public: //---------------------------------------------------------------------
    tileData()
    {
        msb = 0;
        lsb = 0;
    }
//-----------------------------------------------------------------------------
    unsigned int   msb; // might not be very portable
    unsigned int   lsb;
};
//#############################################################################
// @brief this class contains the information about a tile
class tile
{
public: //---------------------------------------------------------------------

    int first;
    int second;
    int third;
    int fourth;
    int height;
    int isVisible;
    int ownerID;
    int tileType;
    int vegetation;
//-----------------------------------------------------------------------------
void toTile(tileData td)
{
    first     =  (td.lsb & 0x0000000F);
    second    = ((td.lsb & 0x000000F0) >> 4);
    third     = ((td.lsb & 0x00000F00) >> 8);
    fourth    = ((td.lsb & 0x0000F000) >> 12);
    height    = ((td.lsb & 0x0FFF0000) >> 16);
    isVisible = ((td.lsb & 0x10000000) >> 28);
 
    ownerID    =  (td.msb & 0x0000003F);
    tileType   = ((td.msb & 0x00000FC0) >> 6);
    vegetation = ((td.msb & 0x0000F000) >> 12);
}
//-----------------------------------------------------------------------------
void toUint(tileData& ft)
{
    ft.lsb  =  (first      & 0x0000000F);
    ft.lsb += ((second     & 0x0000000F) <<  4);
    ft.lsb += ((third      & 0x0000000F) <<  8);
    ft.lsb += ((fourth     & 0x0000000F) << 12);
    ft.lsb += ((height     & 0x00000FFF) << 16);
    ft.lsb += ((isVisible  & 0x00000001) << 28);

    ft.msb =   (ownerID    & 0x0000003F);
    ft.msb += ((tileType   & 0x0000003F) << 6);
    ft.msb += ((vegetation & 0x0000000F) << 12);
}
//-----------------------------------------------------------------------------
bool isOk ()
{
    int minimum = minimal(first,minimal(second,minimal(third,fourth)));
    int maximum = maximal(first,maximal(second,maximal(third,fourth)));
    
    if ((maximum - minimum) > 15) return false;

    if (first  == 0) return true;
    if (second == 0) return true;
    if (third  == 0) return true;
    if (fourth == 0) return true;
    return false;
}
//-----------------------------------------------------------------------------
void insertData(int a, int b, int c, int d)
{
    int minimum = minimal(a,minimal(b,minimal(c,d)));
    int maximum = maximal(a,maximal(b,maximal(c,d)));
    height = minimum;
    first =  a - minimum;
    second = b - minimum;
    third =  c - minimum;
    fourth = d - minimum;
}
//-----------------------------------------------------------------------------
};
//#############################################################################
template <class t>
class  leaf
{
public: //---------------------------------------------------------------------
static tile_size_t   my_lattitude_; // size lattitunal
static tile_size_t   my_longitude_; // size longitunal
//-----------------------------------------------------------------------------
leaf ()
{
    my_data_.resize(my_longitude_ * my_lattitude_);
    setVersion(0);
}
//-----------------------------------------------------------------------------
~leaf()
{
}
//-----------------------------------------------------------------------------
t& operator()(tile_size_t i_long, tile_size_t i_latt)
{
    return my_data_[i_long*my_longitude_ + i_latt];
}
//-----------------------------------------------------------------------------
//tile_size_t latt() { return my_lattitude_; }
////-----------------------------------------------------------------------------
//tile_size_t lngt() { return my_longitude_; }
//-----------------------------------------------------------------------------
void incVersion() { my_version_++; }
//-----------------------------------------------------------------------------
size_t getVersion() { return my_version_; }
//-----------------------------------------------------------------------------
void setVersion(size_t next_ver_ = 0) { my_version_ = next_ver_; };
private: //--------------------------------------------------------------------
    std::vector<t>  my_data_;
    size_t          my_version_;
};
//#############################################################################
template <class t>
class terrain
{
public: //---------------------------------------------------------------------
    terrain(tile_size_t its_longitude, tile_size_t its_lattitude)
{
    leaf<t>::my_lattitude_ = size_factor_;
    leaf<t>::my_longitude_ = size_factor_;
    my_lattitude_ = its_lattitude / size_factor_;
    my_longitude_ = its_longitude / size_factor_;
    my_data_.resize(my_longitude_ * my_lattitude_);
}
//-----------------------------------------------------------------------------
~terrain()
{
}
//-----------------------------------------------------------------------------
void incVersionOf(leaf_size_t i_long, leaf_size_t i_latt)
{
    this->my_data_[i_long*my_longitude_ + i_latt].incVersion();
}
//-----------------------------------------------------------------------------
size_t getVersionOf(leaf_size_t i_long, leaf_size_t i_latt)
{
    return this->my_data_[i_long*my_longitude_ + i_latt].getVersion();
}
//-----------------------------------------------------------------------------
void setVersionOf(leaf_size_t i_long, leaf_size_t i_latt, size_t next_ver_ = 0)
{
    this->my_data_[i_long*my_longitude_ + i_latt].setVersion(next_ver_);
};
//-----------------------------------------------------------------------------
leaf<t>& getLeaf(leaf_size_t i_long, leaf_size_t i_latt)
{
    return this->my_data_[i_long*my_longitude_ + i_latt];
}
//-----------------------------------------------------------------------------
t& operator()(tile_size_t i_long, tile_size_t i_latt)
{
    if (i_latt > (my_lattitude_ * size_factor_))
        i_latt = my_lattitude_ * size_factor_;

    if (i_long > (my_longitude_ * size_factor_))
        i_long = my_longitude_ * size_factor_;

    static size_t reminder_latt = 0;
    static size_t reminder_long = 0;
    static size_t main_latt     = 0;
    static size_t main_long     = 0;

    reminder_latt = i_latt % size_factor_;
    reminder_long = i_long % size_factor_;
    main_latt     = i_latt / size_factor_;
    main_long     = i_long / size_factor_;

    t& retVal = my_data_[main_long*my_longitude_ + main_latt]\
        (reminder_long, reminder_latt);
    return retVal;
}
//-----------------------------------------------------------------------------
size_t latt() { return my_lattitude_; }
//-----------------------------------------------------------------------------
size_t lngt() { return my_longitude_; }
private: //--------------------------------------------------------------------
                   size_t my_longitude_; // longitunal size in leafs
                   size_t my_lattitude_; // lattitunal size in leafs
    std::vector<leaf<t> > my_data_;
public:
             static const size_t size_factor_;
};
//#############################################################################
inline std::ostream& operator <<(std::ostream& a, leaf<tileData> b)
{
    for (int x = 0;x < 8;++x)
    {
        for (int y = 0; y < 8;++y)
        {       
            a << " " << b(x,y).lsb << " ";
        }
        std::cout << "\n";
    }
    return a;
}


//#############################################################################
typedef boost::shared_ptr<terrain<tileData> > terrain_ptr;
//#############################################################################
// this structure is a field of server. It is used to indicate map creation
// task, that a map is supposed to be made.
// the fields are filled when a 'map_make' request was processed by
// a reaction_map_make
struct terrainCreationData
{
    terrainCreationData()
    {
        size_long_  = 0;
        size_latt_  = 0;
        make_map_now_ = false;
    }
    void reset()
    {
        size_long_  = 0;
        size_latt_  = 0;
        make_map_now_ = false;
    }
    tile_size_t size_long_;
    tile_size_t size_latt_;
    volatile bool make_map_now_;
};
//#############################################################################


};
#endif
