/*
This file is a part of TRANSPORT EMPIRE
2006-2007 Copyright by Adam Kadlubek.

This file is licenced under GPL General Public Licence
*/

#ifndef __RENDERER_H__
#define __RENDERER_H__

#include "Ogre.h"
#include "boost/shared_ptr.hpp"
#include "boost/thread/mutex.hpp"
#include "mutexes.h"
#include "map.h"
#include "rendering/render_content_factory.h"
#ifdef WIN32
  #include "timing.h"
#else /* WIN32 */
  #include "timing_posix.h"
#endif
#include <iostream>

extern std::string pluginsName;

namespace te
{

//#############################################################################
class renderer
{
public: //---------------------------------------------------------------------
    /**
    @remark
        constructor
    */
renderer()
{
    my_root_ = boost::shared_ptr<Ogre::Root>(new Ogre::Root(pluginsName));

    Ogre::Root* my_root_raw_ = my_root_.get();

    Ogre::ConfigFile cf;
    cf.load("resources.cfg");

    // Go through all sections & settings in the file
    Ogre::ConfigFile::SectionIterator seci = cf.getSectionIterator();

    Ogre::String secName, typeName, archName;
    while (seci.hasMoreElements())
    {
        secName = seci.peekNextKey();
        Ogre::ConfigFile::SettingsMultiMap *settings = seci.getNext();
        Ogre::ConfigFile::SettingsMultiMap::iterator i;
        for (i = settings->begin(); i != settings->end(); ++i)
        {
            typeName = i->first;
            archName = i->second;
            Ogre::ResourceGroupManager::getSingleton().addResourceLocation(archName, typeName, secName);
        }
    }

    // setup window

    Ogre::RenderSystemList rsl = *(my_root_->getAvailableRenderers());
    Ogre::RenderSystem* renSys = NULL;

    renSys = my_root_->getRenderSystemByName((*rsl.begin())->getName());
    my_root_->setRenderSystem(renSys);

    my_root_->initialise(false);
    my_window_ = my_root_->createRenderWindow("TEmpire", 640, 480, false);

    my_window_->getCustomAttribute("WINDOW",&my_window_handle_);
    my_window_height_ = my_window_->getHeight();
    my_window_width_ = my_window_->getWidth();
  
    // setup scene menager
    my_manager_ = my_root_->createSceneManager(Ogre::ST_GENERIC, "ExampleSMInstance");
 
    // create camera node, will be used for moving the camera around
    cameraRoot = my_manager_->createSceneNode("cameraRoot");
    cameraRoot->setPosition(0,0,0);
    this->my_camera_fov_ = 2.6;
    this->my_camera_latt_pos_ = 0;
    this->my_camera_long_pos_ = 0;
    this->my_inc_camera_fov_ = false;
    this->my_dec_camera_fov_ = false;
    
    // Create the camera
    my_camera_ = my_manager_->createCamera("camera");

    cameraRoot->attachObject(my_camera_);
    my_camera_->setPosition(-500,500,-500);
    my_camera_->lookAt(cameraRoot->getPosition());
    my_camera_->setNearClipDistance(25);
    
    // Create one viewport, entire window
    my_viewport_ = my_window_->addViewport(my_camera_);
    my_viewport_->setBackgroundColour(Ogre::ColourValue(0.0,0.0,0.0));

    // Alter the camera aspect ratio to match the viewport
    my_camera_->setAspectRatio(   Ogre::Real(my_viewport_->getActualWidth())\
                                / Ogre::Real(my_viewport_->getActualHeight())
                               );
    my_camera_->setProjectionType(Ogre::PT_ORTHOGRAPHIC);
    my_camera_->setFOVy(Ogre::Radian(my_camera_fov_));

    // Set default mipmap level (NB some APIs ignore this)
    Ogre::TextureManager::getSingleton().setDefaultNumMipmaps(5);

    // create resource group
    Ogre::ResourceGroupManager::getSingleton().initialiseAllResourceGroups();

    my_manager_->setAmbientLight(Ogre::ColourValue(0.1,0.1,0.1));

    
    my_content_factory_ = boost::shared_ptr<te::render_content_factory>
        (new te::render_content_factory(my_manager_));

    light = my_manager_->createLight("light");

    sunRoot = my_manager_->getRootSceneNode()->createChildSceneNode("sun");
    sunRoot->setPosition(1,-0.75,1);

    light->setSpecularColour(Ogre::ColourValue(1.0,1.0,1.0));
    light->setDiffuseColour(Ogre::ColourValue(1.0,1.0,1.0));
 //   light->setCastShadows(true);
    light->setType(Ogre::Light::LT_DIRECTIONAL);
    //light->setPosition(512,150,512);
    //light->setAttenuation(1000,1.0,0.0,0.0);
    //node2->attachObject(light);

    sunRoot->attachObject(light);

   // my_manager_->setShadowTechnique(Ogre::SHADOWTYPE_STENCIL_ADDITIVE);
    
}
//-----------------------------------------------------------------------------
    /**
    @remark
        dtor
    */
    ~renderer()
    {
    };
//-----------------------------------------------------------------------------
    /**
    @reamrk
        retrieve the numeric handle of the rendering window
    */
    size_t winHandle()
    {
        return my_window_handle_;
    }
//-----------------------------------------------------------------------------
    /**
    @remark
        return height of the rendering window
    */
    size_t winHeight()
    {
        return my_window_height_;
    }
//-----------------------------------------------------------------------------
    /**
    @remark
        return width of the rendering window
    */
    size_t winWidth()
    {
        return my_window_width_;
    }
//-----------------------------------------------------------------------------
    /**
    @remark
        invoke the engine to render one frame
    */
    void operator()()
    {
        lock lk(terrain_operation_);

        static int frame = 0;
  
        static float lastFps;
        static float avgFps;

        Ogre::WindowEventUtilities::messagePump();
        cameraRoot->yaw(Ogre::Radian(0.01));
        sunRoot->yaw(Ogre::Radian(0.01));
        light->setDirection(sunRoot->getPosition());


        if (this->my_dec_camera_fov_)
        {
            this->cameraFovDec();
        }

        if (this->my_inc_camera_fov_)
        {
            this->cameraFovInc();
        }

        my_camera_->setFOVy(Ogre::Radian(my_camera_fov_));

        my_root_->renderOneFrame();
        
        frame++;
 
        if (frame == 96)
        {
            lastFps = this->my_window_->getLastFPS();
            avgFps = this->my_window_->getAverageFPS();
            std::cout << "\nAverage frame rate:" << avgFps << "\nCurrent frame rate:" << lastFps << "\n\n";
            frame = 0;
        }
    }
    //-----------------------------------------------------------------------------
    /**
    @remark
        build a single terrain leaf
    @param
        target - the data of a leaf to draw
    @param
        nodeName - name of a scene node to use to build the leaf
    */
    void buildTerrainLeaf(leaf<tileData>& target,
                          std::string& leafName,
                          leaf_size_t longitude,
                          leaf_size_t lattitude,
                          std::vector<std::string>& children)
    {
        size_t i_x = 0;
        size_t e_x = te::terrain<tileData>::size_factor_;
        size_t i_y = 0;
        size_t e_y = te::terrain<tileData>::size_factor_;
//        render_content_factory::material_params params;

              Ogre::MaterialPtr material;
                  Ogre::Entity* entity = 0;
          Ogre::StaticGeometry* geometry = 0;
                       te::tile tile;
                    std::string entityName;

        //std::cout << "\n---------\n" << "Creating leaf: " << unsigned(longitude) << " " << unsigned(lattitude) << "\n";

        geometry = this->my_manager_->getStaticGeometry(leafName);
        entityName = leafName + "_e";

        int count = 0;

        for (i_x = 0; i_x < e_x; ++i_x)
        {
            for (i_y = 0; i_y < e_y;++i_y)
            {
                ++count;

                // get tile

                tile.toTile(target(i_x,i_y));

                // get node
                
                int vecx = 0;
                int vecy = 0;
                int vecz = 0;
                int angle = 0;

                {
                    lock lk(terrain_operation_);

                    my_manager_->destroyEntity(entityName);
                    entity = this->my_manager_->createEntity(entityName,\
                         my_content_factory_->getTileMesh(tile) + ".tile");

                  //  entity->setCastShadows(true);
                }

                // basic setup done
                if (tile.first == 0)
                {
                    vecx = int(longitude*te::terrain<tileData>::size_factor_+i_x)*32;
                    vecy = int(tile.height);
                    vecz = int(lattitude*te::terrain<tileData>::size_factor_+i_y)*32;
                } else if (tile.second == 0)
                {
                    vecx = int(longitude*te::terrain<tileData>::size_factor_+i_x)*32;
                    vecy = int(tile.height);
                    vecz = int(lattitude*te::terrain<tileData>::size_factor_+i_y)*32+32;
                    angle = 90;
                } else if (tile.third == 0)
                {
                    vecx = int(longitude*te::terrain<tileData>::size_factor_+i_x)*32+32;
                    vecy = int(tile.height);
                    vecz = int(lattitude*te::terrain<tileData>::size_factor_+i_y)*32+32;
                    angle = 180;
                } else if (tile.fourth == 0)
                {
                    vecx = int(longitude*te::terrain<tileData>::size_factor_+i_x)*32+32;
                    vecy = int(tile.height);
                    vecz = int(lattitude*te::terrain<tileData>::size_factor_+i_y)*32;
                    angle = 270;
                }

                {
                    lock lk(terrain_operation_);
                    Ogre::Quaternion quat(Ogre::Radian(Ogre::Degree(angle)),Ogre::Vector3(0,1,0));
                    Ogre::Vector3 vec(vecx,vecy,vecz);
                    if (tile.height < 16)
                    {
                        entity->setMaterialName("te/SeaBottom");
                    } else if (tile.height < 20)
                    {
                        entity->setMaterialName("te/Shore1");
                    } else if (tile.height < 24)
                    {
                        entity->setMaterialName("te/Shore2");
                    }

                    geometry->addEntity(entity,vec,quat);

                    // now if the tile was low enough - add a water level

                    if (tile.height < 32)
                    {
                        my_manager_->destroyEntity(entityName);

                        tile.first = 0;
                        tile.second = 0;
                        tile.third = 0;
                        tile.fourth = 0;
                        tile.height = 20;

                        entity = this->my_manager_->createEntity(entityName,\
                        my_content_factory_->getTileMesh(tile) + ".tile");

                        entity->setMaterialName("te/Water");

                        vecx = int(longitude*te::terrain<tileData>::size_factor_+i_x)*32;
                        vecy = int(tile.height);
                        vecz = int(lattitude*te::terrain<tileData>::size_factor_+i_y)*32;

                        Ogre::Quaternion quat2(Ogre::Radian(0),Ogre::Vector3(0,1,0));
                        vec.x = vecx; vec.y = vecy; vec.z = vecz;
                        geometry->addEntity(entity,vec,quat2);
                    }
                }
            }
        }
    }
//-----------------------------------------------------------------------------
    /**
    @remark
        create a singe instance of terrain geometry which is going to be used
        for this tile
    @param
        name - name of the node
    */
    void createTerrainNode(std::string& OUT leafName,\
              std::vector<std::string>& OUT children)
    {
        lock lk(terrain_operation_);
        static int nn = 0;
        static std::string y("ter_st_g_");
        static std::string nodeName;
        ++nn;
        leafName = y + boost::lexical_cast<std::string>(nn);

        Ogre::StaticGeometry* geom = this->my_manager_->createStaticGeometry(leafName);
    }
//-----------------------------------------------------------------------------
    /**
    @remark
        attach the node to a scene root node
        this will cause the node data to be drawn on next redraw
    @param
        name - name of the node to be attached
    */
    void attachTerrainNode(std::string& leafName)
    {
        lock lk(terrain_operation_);
        my_manager_->getStaticGeometry(leafName)->build();
    }
//-----------------------------------------------------------------------------
    /**
    @remark
        remove node from the root
        this will cause the node objects to be NOT drawn on next
        render, however it will allow modification of the node
    @param
        name - name to be detached
    */
    void detachTerrainNode(std::string& leafName)
    {
        lock lk(terrain_operation_);
        this->my_manager_->getStaticGeometry(leafName)->reset();
    }
//-----------------------------------------------------------------------------
    /**
    @remark
        center camera at certain leaf
    */
    void centerCamera(leaf_size_t longitude, leaf_size_t lattitude)
    {
        cameraRoot->setPosition(longitude*256 + 128,0,lattitude*256 + 128);
    }
//-----------------------------------------------------------------------------
    /**
    * @remark
    *   increase camera fov - ie, zoom out
    */
    void cameraFovInc()
    {
        my_camera_fov_ += 0.005;
        if(my_camera_fov_ > 3.00)
        {
            my_camera_fov_ = 3.00;
        }
    }
//-----------------------------------------------------------------------------
    /**
    * @remark
    *   decrease camera fov - ie, zoom in
    */
    void cameraFovDec()
    {
        my_camera_fov_ -= 0.005;
        if (my_camera_fov_ < 2.5)
        {
            my_camera_fov_ = 2.5;
        }
    }
//-----------------------------------------------------------------------------
    /**
    @remark
        set a flag to keep on increasing camera fov
    */
    void cameraFovInc(bool flag)
    {
        my_inc_camera_fov_ = flag;
    }
//-----------------------------------------------------------------------------
    /**
    @remark
        set a flag to keep on decreasing camera fov
    */
    void cameraFovDec(bool flag)
    {
        my_dec_camera_fov_ = flag;
    }
//-----------------------------------------------------------------------------
    /**
    @remark
        get camera fov
    */
    float getCameraFov()
    {
        return my_camera_fov_;
    }
private: //--------------------------------------------------------------------
    
    boost::shared_ptr<Ogre::Root>   my_root_;
    
                    Ogre::Viewport* my_viewport_;
                Ogre::RenderWindow* my_window_;
                      Ogre::Camera* my_camera_;
                Ogre::SceneManager* my_manager_;
                             size_t my_window_handle_;
                             size_t my_window_width_;
                             size_t my_window_height_;
    boost::shared_ptr<te::render_content_factory>
                                    my_content_factory_;
                              mutex terrain_operation_;
    //- camera -//
                              float my_camera_long_pos_;
                              float my_camera_latt_pos_;
                              float my_camera_fov_;
                               bool my_inc_camera_fov_;
                               bool my_dec_camera_fov_;
                   Ogre::SceneNode* cameraRoot; // camera looks at root

                   Ogre::SceneNode* sunRoot;
                       Ogre::Light* light;
};


};



#endif
