/*
This file is a part of TRANSPORT EMPIRE
2006-2007 Copyright by Adam Kadlubek.

This file is licenced under GPL General Public Licence
*/

/*

This file defines the main processing unit of the TRANSPORT EMPIRE

Server class is the hub of all processing and is responsible for all
synchronisation of game objects and alike.

*/


#ifndef __SERVER_H__
#define __SERVER_H__

#include "exco.h"
#include "task.h"
#include "common/gridArray.h"
#include "action_map.h"
#include "server_actions.h"
#include "server_reactions.h"
#include "communication.h"
#include "boost/shared_ptr.hpp"
#include "boost/utility.hpp"
#include "calendar.h"
#include "map.h"
#include "random.h"

#include <iostream>

namespace te
{
//#############################################################################
class server : public ec::multi_task
{
private:  //-------------------------------------------------------------------
    class processCommands;
    friend class processCommands;

    class tickCalendar;
    friend class tickCalendar;

    class pushGameLogic;
    friend class pushGameLogic;

    class dispatchCommands;
    friend class dispatchCommands;

    class mapMaker;
    friend class mapMaker;

    friend class reaction_map_make;
    friend class reaction_area_req;

public: //---------------------------------------------------------------------
    server(ec::exco& its_list,
           commLink& its_input_comm,
           commLink& its_output_comm,
           size_t userCount,
           terrain_ptr& its_terrain)
     : my_task_list_ (userCount),
       my_in_comm_   (its_input_comm),
       my_out_comm_  (its_output_comm),
       user_count_   (userCount),
       multi_task    (userCount,its_list),
       my_calendar_  (1900,96),
       my_reactions_ (std::string("")),
       my_actions_   (std::string("")),
       my_terrain_   (its_terrain)
{
    my_terrain_ = boost::shared_ptr<terrain<tileData> >\
            (new terrain<tileData>(256,256));

    my_map_creation_data_.make_map_now_ = true;
    my_map_creation_data_.size_latt_ = 256;
    my_map_creation_data_.size_long_ = 256;

    my_task_list_.insert(ec::task_ptr(new processCommands(*this)));
    my_task_list_.insert(ec::task_ptr(new mapMaker(*this)));
    my_task_list_.insert(ec::task_ptr(new tickCalendar(*this)));
    my_task_list_.insert(ec::task_ptr(new pushGameLogic(*this)));
    my_task_list_.insert(ec::task_ptr(new dispatchCommands(*this)));

    my_reactions_.associate(dummy   ,reaction_ptr(new action_dummy()));
    my_reactions_.associate(exit    ,reaction_ptr(new reaction_exit(*this)));
    my_reactions_.associate(map_make,reaction_ptr(new reaction_map_make(*this)));
    my_reactions_.associate(area_req,reaction_ptr(new reaction_area_req(*this)));

    my_actions_.associate(dummy        ,action_ptr(new action_dummy()));
    my_actions_.associate(set_map_size ,action_ptr(new action_set_map_size(my_out_comm_)));
    my_actions_.associate(leaf_send    ,action_ptr(new action_leaf_send(my_out_comm_)));
    my_actions_.associate(date_and_time,action_ptr(new action_date_and_time(my_out_comm_)));

    exit_ = false;
};
//-----------------------------------------------------------------------------
void logic()
    {
        ec::threadlessExecutor tx(my_task_list_,"");
        tx.go();
        my_task_list_.refresh();
        //std::cout << "\nServer Leaf At\n" << my_terrain_->getLeaf(0,3) << "\n";
    }
//-----------------------------------------------------------------------------
bool isExited()
    {
        return this->exit_;
    }
//-----------------------------------------------------------------------------
void bailout()
    {
        exit_ = true;
    }
private: //--------------------------------------------------------------------
    //#########################################################################
    //this task will fetch and process all commands in the input comm queue
    class processCommands : public ec::block_task
    {
    public: //-----------------------------------------------------------------
        processCommands(server& its_server) \
            : ec::block_task (its_server.user_count_, its_server.my_task_list_),
              my_server_     (its_server)
        {
        };
        //---------------------------------------------------------------------
        void logic()
        {
            //std::cout << "server::processCommands\n";
            te::datagram dgram;
            while (my_server_.my_in_comm_.fetch(dgram))
            {
                my_server_.my_reactions_[dgram.command_]->setParams(dgram);
                my_server_.my_reactions_[dgram.command_]->invoke();
            }
        };
    private:
        server& my_server_;
    };
    //#########################################################################
    // this task will increase calendar count
    class tickCalendar : public ec::single_task
    {
    public: //-----------------------------------------------------------------
        tickCalendar(server& its_server)\
            : ec::single_task (its_server.my_task_list_),
              my_server_      (its_server)
        {
        }
    //-------------------------------------------------------------------------
        void logic()
        {
            //std::cout << "server::tickCalendar\n";
            my_server_.my_calendar_.NextDay();
        }
    //-------------------------------------------------------------------------
        server& my_server_;
    };
    //#########################################################################
    // this task is a placeholder for a series of tasks that 
    class pushGameLogic : public ec::multi_task
    {
    public: //-----------------------------------------------------------------
        pushGameLogic(server& its_server)\
            : ec::multi_task (its_server.user_count_, its_server.my_task_list_),
              my_server_     (its_server)
        {
        }
    //-------------------------------------------------------------------------
        void logic()
        {
            //std::cout << "server::PushGameLogic\n";
        }
    //-------------------------------------------------------------------------
        server& my_server_;
    };
    //#########################################################################
    // this task dispatches commands
    class dispatchCommands : public ec::block_task
    {
    public: //-----------------------------------------------------------------
        dispatchCommands(server& its_server)\
            : ec::block_task (its_server.user_count_, its_server.my_task_list_),
              my_server_     (its_server)
        {
        }
    //-------------------------------------------------------------------------
        void logic()
        {
            // insert time send
#define DAT_PTR static_cast<action_date_and_time*> \
    (my_server_.my_actions_[date_and_time].get())
            
            DAT_PTR->setData(my_server_.my_calendar_.getYear());
            DAT_PTR->setData(my_server_.my_calendar_.getMonth());
            DAT_PTR->setData(my_server_.my_calendar_.getDayOfMonth());
            DAT_PTR->setData(my_server_.my_calendar_.getDayOfYear());
            DAT_PTR->invoke();

        }
#undef DAT_PTR
    //-------------------------------------------------------------------------
        server& my_server_;
    };
    //#########################################################################
    class mapMaker : public ec::multi_task
    {
    public: //-----------------------------------------------------------------
        mapMaker(server& its_server)\
            : ec::multi_task(its_server.user_count_, its_server.my_task_list_),
              my_server_    (its_server),
              my_barr_      (static_cast<unsigned int>(its_server.user_count_))
        {
        }
    //-------------------------------------------------------------------------
        ~mapMaker()
        {
        }
    //-------------------------------------------------------------------------
        void logic()
        {
            if (my_server_.my_map_creation_data_.make_map_now_)
            {
                // AI convert all the code below to be reentrant
                if (my_barr_.wait())
                {

                    ulong longitude  = (ulong)my_server_.my_map_creation_data_.size_long_;
                    ulong lattitude  = (ulong)my_server_.my_map_creation_data_.size_latt_;
                    gridArrayPtr array = 0;
                    makeArray(longitude/2,lattitude/2,randomfn,&array);
                    go(4,&array);
                    scale(longitude+1, lattitude+1, &array);
                    filter(3,&array);
                    constrainMap(19,48,7,array);
                    filter(5,&array);
 
                    my_server_.my_terrain_ = \
                        boost::shared_ptr<terrain<tileData> >\
                        (new terrain<tileData>(longitude, lattitude));

                    tile     converter;
                    tileData inserter;

                    for (ulong x = 0;x < longitude;++x)
                    {
                        std::cout << ".";
                        for (ulong y = 0; y < lattitude;++y)
                        {
                            converter.insertData
                                (
                                int(val(x  ,y  ,array)),
                                int(val(x  ,y+1,array)),
                                int(val(x+1,y+1,array)),
                                int(val(x+1,y  ,array))
                                );

                                if (converter.isOk())
                                {
                                    converter.toUint((*my_server_.my_terrain_)(x,y));
                                    (*my_server_.my_terrain_).setVersionOf(x/8,y/8,1);
                                } else {
                                    throw 1;
                                }
                        }
                    }
                    destroyArray(&array);
                    my_server_.my_map_creation_data_.reset();

#define ASMS_PTR static_cast<action_set_map_size*> \
    (my_server_.my_actions_[set_map_size].get())

                    ASMS_PTR->setSize(longitude, lattitude);
                    my_server_.my_actions_[set_map_size]->invoke();

#undef ASMS_PTR
                }
                my_barr_.wait();
            }
        }
    //-------------------------------------------------------------------------
        server& my_server_;
        boost::barrier my_barr_;
    };
    //#########################################################################
private: //--------------------------------------------------------------------
                                        commLink& my_in_comm_;
                                        commLink& my_out_comm_;
                                         ec::exco my_task_list_;
                                           size_t user_count_;
                                     te::Calendar my_calendar_;
                                 te::reaction_map my_reactions_;
                                   te::action_map my_actions_;
                                    volatile bool exit_;
                              terrainCreationData my_map_creation_data_;
                                     terrain_ptr& my_terrain_;
};
//#############################################################################
};


#endif
